/**
   Implementazione di un server TCP che risponde al comando 'PING'
   ritornando la stringa 'PONG' e poi termina.
 **/

#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#define BUFFSIZE 1024
#define PORT 1337

const char *COMMAND_STRING = "PING";
const char *OUTPUT_STRING = "PONG\n";

int main() {
  int sockfd;
  int connfd;
  int tmp;
  unsigned int len;
  ssize_t msglen;
  char buf[BUFFSIZE];
  struct sockaddr_in serveraddr;
  struct sockaddr_in client;

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0) {
    fprintf(stderr, "ERROR: socket() call failed\n");
    return 1;
  }

  // enable socket options
  const int enable = 1;
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0) {
    fprintf(stderr, "ERROR: setsockopt(SO_REUSEADDR) failed\n");
    return 1;    
  }

  // fill server addr struct
  bzero(&serveraddr, sizeof(serveraddr));         // initialize to 0
  serveraddr.sin_family = AF_INET;
  serveraddr.sin_addr.s_addr = htonl(INADDR_ANY); // host-to-network long
  serveraddr.sin_port = htons(PORT);              // host-to-network short

  if (bind(sockfd, (struct sockaddr *)&serveraddr, sizeof(serveraddr)) < 0) {
    fprintf(stderr, "ERROR: bind() call failed\n");
    if (errno != 0) {
      fprintf(stderr, "%s\n", strerror(errno));
    }
    return 1;
  }

  // listen for incoming connections
  printf("Listening on port %d!\n", PORT);
  // backlog set to 1
  if (listen(sockfd, 1) != 0) {
    fprintf(stderr, "ERROR: listen() call failed");
    return 1;    
  }

  // accept a connection from a client
  len = sizeof(client);
  connfd = accept(sockfd, (struct sockaddr *)&client, &len);
  if (connfd == -1) {
    fprintf(stderr, "ERROR: accept() call failed");
    return 1;
  }
  // receive message
  msglen = recv(connfd, buf, BUFFSIZE, 0);
  if (msglen == -1)  {
    fprintf(stderr, "ERROR: receive() call failed");
    return 1;
  }

  // check if the message is valid
  tmp = strncmp(buf, COMMAND_STRING, strlen(COMMAND_STRING));

  if (tmp != 0) {
    printf("Invalid command sent by the client!\n");
    char *error_msg = "ERROR: Invalid command!\n";
    msglen = strlen(error_msg) + 1;
    strncpy(buf, error_msg, msglen);
  } else {
    printf("OK – Valid command sent by the client!\n");
    printf("OK – About to send correct response\n");
    msglen = strlen(OUTPUT_STRING) + 1;
    strncpy(buf, OUTPUT_STRING, msglen);
  }

  // send message back to the client
  if (send(connfd, buf, msglen, 0) < 0) {
    fprintf(stderr, "ERROR: send() call failed");
    return 1;
  }

  close(connfd);
  close(sockfd);
  
  return 0;
}
